"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
var assignment_1 = require("./assignment");
var ts = require("typescript");
var template_1 = require("../template");
var types_1 = require("../types");
var variable_1 = require("./variable");
var regexfunc_1 = require("./regexfunc");
var literals_1 = require("./literals");
var typeconvert_1 = require("./typeconvert");
var typeguards_1 = require("../typeguards");
var elementaccess_1 = require("./elementaccess");
var standard_1 = require("../standard");
var CCondition = /** @class */ (function () {
    function CCondition(scope, node) {
        this.universalWrapper = false;
        this.isString = false;
        this.expressionIsIdentifier = false;
        this.expression = template_1.CodeTemplateFactory.createForNode(scope, node);
        this.expressionIsIdentifier = ts.isIdentifier(node);
        var type = scope.root.typeHelper.getCType(node);
        this.isString = type == types_1.StringVarType;
        if (type == types_1.UniversalVarType) {
            this.universalWrapper = true;
            scope.root.headerFlags.js_var_to_bool = true;
        }
    }
    CCondition = __decorate([
        template_1.CodeTemplate("\n{#if universalWrapper}\n    js_var_to_bool({expression})\n{#elseif isString && expressionIsIdentifier}\n    *{expression}\n{#elseif isString}\n    *({expression})\n{#else}\n    {expression}\n{/if}")
    ], CCondition);
    return CCondition;
}());
exports.CCondition = CCondition;
var CBinaryExpression = /** @class */ (function () {
    function CBinaryExpression(scope, node) {
        this.expression = null;
        if (node.operatorToken.kind == ts.SyntaxKind.EqualsToken) {
            this.expression = assignment_1.AssignmentHelper.create(scope, node.left, node.right, true);
            return;
        }
        if (node.operatorToken.kind == ts.SyntaxKind.CommaToken) {
            var nodeAsStatement = ts.createNode(ts.SyntaxKind.ExpressionStatement);
            nodeAsStatement.expression = node.left;
            nodeAsStatement.parent = node.getSourceFile();
            scope.statements.push(template_1.CodeTemplateFactory.createForNode(scope, nodeAsStatement));
            this.expression = template_1.CodeTemplateFactory.createForNode(scope, node.right);
            return;
        }
        if (node.operatorToken.kind == ts.SyntaxKind.PlusToken) {
            this.expression = new CPlusExpression(scope, node);
            return;
        }
        if (node.operatorToken.kind == ts.SyntaxKind.PlusEqualsToken) {
            var left = template_1.CodeTemplateFactory.createForNode(scope, node.left);
            var right = new CPlusExpression(scope, node);
            this.expression = "(" + template_1.CodeTemplateFactory.templateToString(left) + " = " + template_1.CodeTemplateFactory.templateToString(right) + ")";
        }
        if (typeguards_1.isNumberOp(node.operatorToken.kind) || typeguards_1.isIntegerOp(node.operatorToken.kind)) {
            this.expression = new CArithmeticExpression(scope, node);
            return;
        }
        if (typeguards_1.isRelationalOp(node.operatorToken.kind)) {
            this.expression = new CRelationalExpression(scope, node);
            return;
        }
        if (typeguards_1.isEqualityOp(node.operatorToken.kind)) {
            this.expression = new CEqualityExpression(scope, node);
            return;
        }
        if (node.operatorToken.kind === ts.SyntaxKind.InKeyword) {
            this.expression = new CInExpression(scope, node);
            return;
        }
        if (typeguards_1.isLogicOp(node.operatorToken.kind)) {
            this.expression = new CLogicExpession(scope, node);
            return;
        }
        this.nodeText = node.flags & ts.NodeFlags.Synthesized ? "(synthesized node)" : node.getText();
    }
    CBinaryExpression = __decorate([
        template_1.CodeTemplate("\n{#if expression}\n    {expression}\n{#else}\n    /* unsupported expression {nodeText} */\n{/if}", ts.SyntaxKind.BinaryExpression)
    ], CBinaryExpression);
    return CBinaryExpression;
}());
exports.CBinaryExpression = CBinaryExpression;
var CLogicExpession = /** @class */ (function () {
    function CLogicExpession(scope, node) {
        this.leftVarName = "";
        this.rightVarName = "";
        var type = scope.root.typeHelper.getCType(node);
        if (type === types_1.UniversalVarType) {
            this.left = new typeconvert_1.CAsUniversalVar(scope, node.left);
            this.right = new typeconvert_1.CAsUniversalVar(scope, node.right);
        }
        else {
            this.left = template_1.CodeTemplateFactory.createForNode(scope, node.left);
            this.right = template_1.CodeTemplateFactory.createForNode(scope, node.right);
        }
        this.isBoolContext = typeguards_1.isInBoolContext(node) && type !== types_1.UniversalVarType;
        var isOr = node.operatorToken.kind === ts.SyntaxKind.BarBarToken;
        if (this.isBoolContext) {
            this.operator = isOr ? "||" : "&&";
        }
        else {
            if (!typeguards_1.isSimpleNode(node.left)) {
                this.leftVarName = scope.root.symbolsHelper.addTemp(node, "tmp1");
                scope.variables.push(new variable_1.CVariable(scope, this.leftVarName, type));
            }
            if (!typeguards_1.isSimpleNode(node.right)) {
                this.rightVarName = scope.root.symbolsHelper.addTemp(node, "tmp2");
                scope.variables.push(new variable_1.CVariable(scope, this.rightVarName, type));
            }
            if (this.leftVarName && type === types_1.UniversalVarType) {
                this.condition = "js_var_to_bool(" + this.leftVarName + ")";
                scope.root.headerFlags.js_var_to_bool = true;
            }
            else
                this.condition = this.leftVarName || new CCondition(scope, node.left);
            if (isOr) {
                this.whenTrue = this.leftVarName || this.left;
                this.whenFalse = this.rightVarName || this.right;
            }
            else {
                this.whenTrue = this.rightVarName || this.right;
                this.whenFalse = this.leftVarName || this.left;
            }
        }
    }
    CLogicExpession = __decorate([
        template_1.CodeTemplate("\n{#statements}\n    {#if leftVarName}\n        {leftVarName} = {left};\n    {/if}\n    {#if rightVarName}\n        {rightVarName} = {right};\n    {/if}\n{/statements}\n{#if isBoolContext}\n    {left} {operator} {right}\n{#else}\n    {condition} ? {whenTrue} : {whenFalse}\n{/if}")
    ], CLogicExpession);
    return CLogicExpession;
}());
var CArithmeticExpression = /** @class */ (function () {
    function CArithmeticExpression(scope, node) {
        this.operator = null;
        this.computeOperation = null;
        var leftType = scope.root.typeHelper.getCType(node.left);
        var rightType = scope.root.typeHelper.getCType(node.right);
        this.isCompoundAssignment = typeguards_1.isCompoundAssignment(node.operatorToken);
        if (types_1.toNumberCanBeNaN(leftType) || types_1.toNumberCanBeNaN(rightType)) {
            var js_var_operator_map = (_a = {},
                _a[ts.SyntaxKind.AsteriskToken] = "JS_VAR_ASTERISK",
                _a[ts.SyntaxKind.AsteriskEqualsToken] = "JS_VAR_ASTERISK",
                _a[ts.SyntaxKind.SlashToken] = "JS_VAR_SLASH",
                _a[ts.SyntaxKind.SlashEqualsToken] = "JS_VAR_SLASH",
                _a[ts.SyntaxKind.PercentToken] = "JS_VAR_PERCENT",
                _a[ts.SyntaxKind.PercentEqualsToken] = "JS_VAR_PERCENT",
                _a[ts.SyntaxKind.MinusToken] = "JS_VAR_MINUS",
                _a[ts.SyntaxKind.MinusEqualsToken] = "JS_VAR_MINUS",
                _a[ts.SyntaxKind.LessThanLessThanToken] = "JS_VAR_SHL",
                _a[ts.SyntaxKind.LessThanLessThanEqualsToken] = "JS_VAR_SHL",
                _a[ts.SyntaxKind.GreaterThanGreaterThanToken] = "JS_VAR_SHR",
                _a[ts.SyntaxKind.GreaterThanGreaterThanEqualsToken] = "JS_VAR_SHR",
                _a[ts.SyntaxKind.GreaterThanGreaterThanGreaterThanToken] = "JS_VAR_USHR",
                _a[ts.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken] = "JS_VAR_USHR",
                _a[ts.SyntaxKind.BarToken] = "JS_VAR_OR",
                _a[ts.SyntaxKind.BarEqualsToken] = "JS_VAR_OR",
                _a[ts.SyntaxKind.AmpersandToken] = "JS_VAR_AND",
                _a[ts.SyntaxKind.AmpersandEqualsToken] = "JS_VAR_AND",
                _a);
            this.computeOperation = js_var_operator_map[node.operatorToken.kind];
            this.left = new typeconvert_1.CAsUniversalVar(scope, node.left);
            this.right = new typeconvert_1.CAsUniversalVar(scope, node.right);
            scope.root.headerFlags.js_var_compute = true;
        }
        else {
            this.operator = node.operatorToken.getText();
            this.left = new typeconvert_1.CAsNumber(scope, node.left);
            this.right = new typeconvert_1.CAsNumber(scope, node.right);
            if (node.operatorToken.kind == ts.SyntaxKind.GreaterThanGreaterThanGreaterThanToken
                || node.operatorToken.kind == ts.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken) {
                this.operator = ">>";
                var leftAsString = template_1.CodeTemplateFactory.templateToString(this.left);
                this.left = "((uint16_t)" + leftAsString + ")";
                if (node.operatorToken.kind == ts.SyntaxKind.GreaterThanGreaterThanGreaterThanEqualsToken)
                    this.left = leftAsString + " = " + this.left;
                scope.root.headerFlags.uint16_t = true;
            }
        }
        this.nodeText = node.flags & ts.NodeFlags.Synthesized ? "(synthesized node)" : node.getText();
        var _a;
    }
    CArithmeticExpression = __decorate([
        template_1.CodeTemplate("\n{#if operator}\n    {left} {operator} {right}\n{#elseif computeOperation && isCompoundAssignment}\n    {left} = js_var_compute({left}, {computeOperation}, {right})\n{#elseif computeOperation}\n    js_var_compute({left}, {computeOperation}, {right})\n{#else}\n    /* unsupported arithmetic expression {nodeText} */\n{/if}")
    ], CArithmeticExpression);
    return CArithmeticExpression;
}());
var CRelationalExpression = /** @class */ (function () {
    function CRelationalExpression(scope, node) {
        this.operator = null;
        this.universalCondition = null;
        this.stringCondition = null;
        var leftType = scope.root.typeHelper.getCType(node.left);
        var rightType = scope.root.typeHelper.getCType(node.right);
        if (leftType === types_1.UniversalVarType || rightType === types_1.UniversalVarType) {
            switch (node.operatorToken.kind) {
                case ts.SyntaxKind.LessThanToken:
                    this.left = new typeconvert_1.CAsUniversalVar(scope, node.left);
                    this.right = new typeconvert_1.CAsUniversalVar(scope, node.right);
                    this.universalCondition = "> 0";
                    break;
                case ts.SyntaxKind.LessThanEqualsToken:
                    // notice operands are swapped
                    this.left = new typeconvert_1.CAsUniversalVar(scope, node.right);
                    this.right = new typeconvert_1.CAsUniversalVar(scope, node.left);
                    this.universalCondition = "< 0";
                    break;
                case ts.SyntaxKind.GreaterThanToken:
                    // notice operands are swapped
                    this.left = new typeconvert_1.CAsUniversalVar(scope, node.right);
                    this.right = new typeconvert_1.CAsUniversalVar(scope, node.left);
                    this.universalCondition = "> 0";
                    break;
                case ts.SyntaxKind.GreaterThanEqualsToken:
                    this.left = new typeconvert_1.CAsUniversalVar(scope, node.left);
                    this.right = new typeconvert_1.CAsUniversalVar(scope, node.right);
                    this.universalCondition = "< 0";
                    break;
            }
            scope.root.headerFlags.js_var_lessthan = true;
        }
        else if (leftType === types_1.StringVarType && rightType === types_1.StringVarType) {
            this.stringCondition = node.operatorToken.getText() + " 0";
            this.left = template_1.CodeTemplateFactory.createForNode(scope, node.left);
            this.right = template_1.CodeTemplateFactory.createForNode(scope, node.right);
            scope.root.headerFlags.strings = true;
        }
        else {
            this.operator = node.operatorToken.getText();
            this.left = new typeconvert_1.CAsNumber(scope, node.left);
            this.right = new typeconvert_1.CAsNumber(scope, node.right);
        }
        this.nodeText = node.flags & ts.NodeFlags.Synthesized ? "(synthesized node)" : node.getText();
    }
    CRelationalExpression = __decorate([
        template_1.CodeTemplate("\n{#if operator}\n    {left} {operator} {right}\n{#elseif stringCondition}\n    strcmp({left}, {right}) {stringCondition}\n{#elseif universalCondition}\n    js_var_lessthan({left}, {right}) {universalCondition}\n{#else}\n    /* unsupported relational expression {nodeText} */\n{/if}")
    ], CRelationalExpression);
    return CRelationalExpression;
}());
var CEqualityExpression = /** @class */ (function () {
    function CEqualityExpression(scope, node) {
        this.expression = null;
        this.operator = null;
        this.stringCondition = null;
        this.strNumCondition = null;
        this.universalCondition = null;
        this.strict = null;
        var leftType = scope.root.typeHelper.getCType(node.left);
        var rightType = scope.root.typeHelper.getCType(node.right);
        var notEquals = node.operatorToken.kind === ts.SyntaxKind.ExclamationEqualsEqualsToken || node.operatorToken.kind === ts.SyntaxKind.ExclamationEqualsToken;
        this.strict = node.operatorToken.kind === ts.SyntaxKind.ExclamationEqualsEqualsToken || node.operatorToken.kind === ts.SyntaxKind.EqualsEqualsEqualsToken ? "TRUE" : "FALSE";
        this.left = template_1.CodeTemplateFactory.createForNode(scope, node.left);
        this.right = template_1.CodeTemplateFactory.createForNode(scope, node.right);
        if ((leftType == types_1.NumberVarType || leftType == types_1.BooleanVarType) && (rightType == types_1.NumberVarType || rightType == types_1.BooleanVarType)) {
            this.operator = notEquals ? "!=" : "==";
        }
        else if (leftType == types_1.StringVarType && rightType == types_1.StringVarType) {
            this.stringCondition = notEquals ? "!= 0" : "== 0";
            scope.root.headerFlags.strings = true;
        }
        else if (leftType == types_1.NumberVarType && rightType == types_1.StringVarType
            || leftType == types_1.StringVarType && rightType == types_1.NumberVarType) {
            this.strNumCondition = notEquals ? "!= 0" : "== 0";
            scope.root.headerFlags.str_int16_t_cmp = true;
            // str_int16_t_cmp expects certain order of arguments (string, number)
            if (leftType == types_1.NumberVarType) {
                var tmp = this.left;
                this.left = this.right;
                this.right = tmp;
            }
        }
        else if (leftType == types_1.UniversalVarType || rightType == types_1.UniversalVarType) {
            this.universalCondition = notEquals ? "== FALSE" : "== TRUE";
            this.left = new typeconvert_1.CAsUniversalVar(scope, this.left, leftType);
            this.right = new typeconvert_1.CAsUniversalVar(scope, this.right, rightType);
            scope.root.headerFlags.js_var_eq = true;
        }
        else if (leftType instanceof types_1.StructType || leftType instanceof types_1.ArrayType || leftType instanceof types_1.DictType
            || rightType instanceof types_1.StructType || rightType instanceof types_1.ArrayType || rightType instanceof types_1.DictType) {
            if (leftType != rightType) {
                this.expression = notEquals ? "TRUE" : "FALSE";
                scope.root.headerFlags.bool = true;
            }
            else
                this.operator = notEquals ? "!=" : "==";
        }
        this.nodeText = node.flags & ts.NodeFlags.Synthesized ? "(synthesized node)" : node.getText();
    }
    CEqualityExpression = __decorate([
        template_1.CodeTemplate("\n{#if expression}\n    {expression}\n{#elseif operator}\n    {left} {operator} {right}\n{#elseif stringCondition}\n    strcmp({left}, {right}) {stringCondition}\n{#elseif strNumCondition}\n    str_int16_t_cmp({left}, {right}) {strNumCondition}\n{#elseif universalCondition}\n    js_var_eq({left}, {right}, {strict}) {universalCondition}\n{#else}\n    /* unsupported equality expression {nodeText} */\n{/if}")
    ], CEqualityExpression);
    return CEqualityExpression;
}());
var CPlusExpression = /** @class */ (function () {
    function CPlusExpression(scope, node) {
        this.addNumbers = false;
        this.isUniversalVar = false;
        this.replacedWithVar = false;
        this.replacementVarName = null;
        this.gcVarName = null;
        var leftType = scope.root.typeHelper.getCType(node.left);
        this.left = template_1.CodeTemplateFactory.createForNode(scope, node.left);
        var rightType = scope.root.typeHelper.getCType(node.right);
        this.right = template_1.CodeTemplateFactory.createForNode(scope, node.right);
        if (leftType == types_1.RegexVarType) {
            leftType = types_1.StringVarType;
            this.left = new regexfunc_1.CRegexAsString(this.left);
        }
        if (rightType == types_1.RegexVarType) {
            rightType = types_1.StringVarType;
            this.right = new regexfunc_1.CRegexAsString(this.right);
        }
        if ((leftType === types_1.NumberVarType || leftType === types_1.BooleanVarType) && (rightType === types_1.NumberVarType || rightType === types_1.BooleanVarType)) {
            this.addNumbers = true;
        }
        else if (leftType === types_1.UniversalVarType || rightType === types_1.UniversalVarType) {
            this.isUniversalVar = true;
            this.left = new typeconvert_1.CAsUniversalVar(scope, this.left, leftType);
            this.right = new typeconvert_1.CAsUniversalVar(scope, this.right, rightType);
            scope.root.headerFlags.js_var_plus = true;
        }
        else {
            var tempVarName = scope.root.memoryManager.getReservedTemporaryVarName(node);
            scope.func.variables.push(new variable_1.CVariable(scope, tempVarName, "char *", { initializer: "NULL" }));
            this.gcVarName = scope.root.memoryManager.getGCVariableForNode(node);
            this.replacedWithVar = true;
            this.replacementVarName = tempVarName;
            this.strlen_left = new typeconvert_1.CAsString_Length(scope, node.left, this.left, leftType);
            this.strlen_right = new typeconvert_1.CAsString_Length(scope, node.right, this.right, rightType);
            this.strcat_left = new typeconvert_1.CAsString_Concat(scope, node.left, tempVarName, this.left, leftType);
            this.strcat_right = new typeconvert_1.CAsString_Concat(scope, node.right, tempVarName, this.right, rightType);
            scope.root.headerFlags.strings = true;
            scope.root.headerFlags.malloc = true;
            scope.root.headerFlags.str_int16_t_cat = true;
            if (this.gcVarName) {
                scope.root.headerFlags.gc_iterator = true;
                scope.root.headerFlags.array = true;
            }
        }
    }
    CPlusExpression = __decorate([
        template_1.CodeTemplate("\n{#statements}\n    {#if replacedWithVar}\n        {replacementVarName} = malloc({strlen_left} + {strlen_right} + 1);\n        assert({replacementVarName} != NULL);\n        {replacementVarName}[0] = '\\0';\n        {strcat_left}\n        {strcat_right}\n    {/if}\n    {#if replacedWithVar && gcVarName}\n        ARRAY_PUSH({gcVarName}, {replacementVarName});\n    {/if}\n\n{/statements}\n{#if addNumbers}\n    {left} + {right}\n{#elseif replacedWithVar}\n    {replacementVarName}\n{#elseif isUniversalVar}\n    js_var_plus({left}, {right})\n{/if}")
    ], CPlusExpression);
    return CPlusExpression;
}());
var CInExpression = /** @class */ (function () {
    function CInExpression(scope, node) {
        var _this = this;
        this.isArray = false;
        this.isStruct = false;
        this.isDict = false;
        this.isUniversalVar = false;
        this.result = null;
        this.tmpVarName = null;
        var type = scope.root.typeHelper.getCType(node.right);
        this.obj = template_1.CodeTemplateFactory.createForNode(scope, node.right);
        if (type instanceof types_1.ArrayType) {
            this.isArray = true;
            this.arraySize = new elementaccess_1.CArraySize(scope, this.obj, type);
            this.key = new typeconvert_1.CAsNumber(scope, node.left);
            var keyType = scope.root.typeHelper.getCType(node.left);
            if (types_1.toNumberCanBeNaN(keyType)) {
                this.tmpVarName = scope.root.symbolsHelper.addTemp(node, "tmp_key");
                scope.variables.push(new variable_1.CVariable(scope, this.tmpVarName, types_1.UniversalVarType));
            }
        }
        else {
            this.key = new typeconvert_1.CAsString(scope, node.left);
        }
        if (type instanceof types_1.StructType) {
            this.isStruct = true;
            var propTypes = Object.keys(type.properties);
            if (propTypes.length == 0) {
                this.result = "FALSE";
                scope.root.headerFlags.bool = true;
            }
            else {
                var initializer = "{ " + propTypes.sort().map(function (p) { return '"' + p + '"'; }).join(", ") + " }";
                this.propertiesVarName = type.structName + "_props";
                this.propertiesCount = propTypes.length + "";
                if (!scope.root.variables.some(function (v) { return v.name === _this.propertiesVarName; }))
                    scope.root.variables.push(new variable_1.CVariable(scope, this.propertiesVarName, "const char *{var}[" + this.propertiesCount + "]", { initializer: initializer }));
                scope.root.headerFlags.dict_find_pos = true;
            }
        }
        this.isDict = type instanceof types_1.DictType;
        this.isUniversalVar = type === types_1.UniversalVarType;
        if (ts.isStringLiteral(node.left)) {
            var ident_1 = ts.createIdentifier(node.left.text);
            var propAccess_1 = ts.createPropertyAccess(node.right, ident_1);
            var standardCall = ts.createCall(propAccess_1, [], []);
            ident_1.parent = propAccess_1;
            ident_1.getText = function () { return ident_1.text; };
            propAccess_1.parent = standardCall;
            propAccess_1.getText = function () { return "(" + node.right.getText() + ")." + ident_1.text; };
            standardCall.parent = node.parent;
            standardCall.getText = function () { return propAccess_1.getText() + "()"; };
            if (standard_1.StandardCallHelper.isStandardCall(scope.root.typeHelper, standardCall))
                this.result = "TRUE";
        }
        if (this.isArray && ts.isStringLiteral(node.left) && node.left.text === "length")
            this.result = "TRUE";
        this.nodeText = node.getText();
    }
    CInExpression = __decorate([
        template_1.CodeTemplate("\n{#statements}\n    {#if tmpVarName}\n        {tmpVarName} = {key};\n    {/if}\n{/statements}\n{#if result}\n    {result}\n{#elseif isArray && tmpVarName}\n    ({tmpVarName}.type != JS_VAR_NAN && {tmpVarName}.number >= 0 && {tmpVarName}.number < {arraySize})\n{#elseif isArray && !tmpVarName}\n    ({key} >= 0 && {key} < {arraySize})\n{#elseif isStruct}\n    dict_find_pos({propertiesVarName}, {propertiesCount}, {key}) > -1\n{#elseif isDict}\n    dict_find_pos({obj}->index->data, {obj}->index->size, {key}) > -1\n{#elseif isUniversalVar}\n    js_var_get({obj}, {key}).type != JS_VAR_UNDEFINED\n{#else}\n    /* unsupported 'in' expression {nodeText} */\n{/if}")
    ], CInExpression);
    return CInExpression;
}());
var CUnaryExpression = /** @class */ (function () {
    function CUnaryExpression(scope, node) {
        this.before = "";
        this.after = "";
        this.isCompound = false;
        this.isPostfix = node.kind == ts.SyntaxKind.PostfixUnaryExpression;
        var type = scope.root.typeHelper.getCType(node.operand);
        if (node.operator === ts.SyntaxKind.PlusToken)
            this.operand = new typeconvert_1.CAsNumber(scope, node.operand);
        else if (node.operator === ts.SyntaxKind.MinusToken) {
            this.before = "-";
            this.operand = new typeconvert_1.CAsNumber(scope, node.operand);
            if (types_1.toNumberCanBeNaN(type)) {
                this.before = "js_var_compute(js_var_from_int16_t(0), JS_VAR_MINUS, ";
                this.after = ")";
                scope.root.headerFlags.js_var_compute = true;
                scope.root.headerFlags.js_var_from_int16_t = true;
            }
        }
        else if (node.operator === ts.SyntaxKind.TildeToken) {
            this.before = "~";
            this.operand = new typeconvert_1.CAsNumber(scope, node.operand);
            if (types_1.toNumberCanBeNaN(type))
                this.after = ".number";
        }
        else if (node.operator === ts.SyntaxKind.ExclamationToken) {
            this.before = "!";
            this.operand = new CCondition(scope, node.operand);
        }
        else if (node.operator === ts.SyntaxKind.PlusPlusToken) {
            if (this.isPostfix) {
                if (types_1.toNumberCanBeNaN(type))
                    this.operand = "/* not supported expression " + node.getText() + " */";
                else {
                    this.operand = new typeconvert_1.CAsNumber(scope, node.operand);
                    this.after = "++";
                }
            }
            else {
                if (types_1.toNumberCanBeNaN(type)) {
                    this.isCompound = true;
                    this.before = "js_var_plus(js_var_to_number(";
                    this.operand = template_1.CodeTemplateFactory.createForNode(scope, node.operand);
                    this.after = "), js_var_from_int16_t(1))";
                    scope.root.headerFlags.js_var_plus = true;
                    scope.root.headerFlags.js_var_from_int16_t = true;
                }
                else {
                    this.before = "++";
                    this.operand = new typeconvert_1.CAsNumber(scope, node.operand);
                }
            }
        }
        else if (node.operator === ts.SyntaxKind.MinusMinusToken) {
            if (this.isPostfix) {
                if (types_1.toNumberCanBeNaN(type))
                    this.operand = "/* not supported expression " + node.getText() + " */";
                else {
                    this.operand = new typeconvert_1.CAsNumber(scope, node.operand);
                    this.after = "--";
                }
            }
            else {
                if (types_1.toNumberCanBeNaN(type)) {
                    this.isCompound = true;
                    this.before = "js_var_compute(";
                    this.operand = template_1.CodeTemplateFactory.createForNode(scope, node.operand);
                    this.after = ", JS_VAR_MINUS, js_var_from_int16_t(1))";
                    scope.root.headerFlags.js_var_compute = true;
                    scope.root.headerFlags.js_var_from_int16_t = true;
                }
                else {
                    this.before = "--";
                    this.operand = new typeconvert_1.CAsNumber(scope, node.operand);
                }
            }
        }
    }
    CUnaryExpression = __decorate([
        template_1.CodeTemplate("\n{#if isCompound}\n    ({operand} = {before}{operand}{after})\n{#else}\n    {before}{operand}{after}\n{/if}", [ts.SyntaxKind.PrefixUnaryExpression, ts.SyntaxKind.PostfixUnaryExpression])
    ], CUnaryExpression);
    return CUnaryExpression;
}());
var CTernaryExpression = /** @class */ (function () {
    function CTernaryExpression(scope, node) {
        this.condition = template_1.CodeTemplateFactory.createForNode(scope, node.condition);
        this.whenTrue = template_1.CodeTemplateFactory.createForNode(scope, node.whenTrue);
        this.whenFalse = template_1.CodeTemplateFactory.createForNode(scope, node.whenFalse);
    }
    CTernaryExpression = __decorate([
        template_1.CodeTemplate("{condition} ? {whenTrue} : {whenFalse}", ts.SyntaxKind.ConditionalExpression)
    ], CTernaryExpression);
    return CTernaryExpression;
}());
var CGroupingExpression = /** @class */ (function () {
    function CGroupingExpression(scope, node) {
        this.expression = template_1.CodeTemplateFactory.createForNode(scope, node.expression);
    }
    CGroupingExpression = __decorate([
        template_1.CodeTemplate("({expression})", ts.SyntaxKind.ParenthesizedExpression)
    ], CGroupingExpression);
    return CGroupingExpression;
}());
var CTypeOf = /** @class */ (function () {
    function CTypeOf(scope, node) {
        var type = scope.root.typeHelper.getCType(node.expression);
        this.isUniversalVar = type === types_1.UniversalVarType;
        this.isString = type === types_1.StringVarType;
        this.isNumber = type === types_1.NumberVarType;
        this.isBoolean = type === types_1.BooleanVarType;
        this.expression = template_1.CodeTemplateFactory.createForNode(scope, node.expression);
        if (type == types_1.UniversalVarType) {
            scope.root.headerFlags.js_var = true;
            scope.root.headerFlags.js_var_typeof = true;
        }
    }
    CTypeOf = __decorate([
        template_1.CodeTemplate("\n{#if isUniversalVar}\n    js_var_typeof({expression})\n{#elseif isString}\n    \"string\"\n{#elseif isNumber}\n    \"number\"\n{#elseif isBoolean}\n    \"number\"\n{#else}\n    \"object\"\n{/if}", ts.SyntaxKind.TypeOfExpression)
    ], CTypeOf);
    return CTypeOf;
}());
var CVoid = /** @class */ (function () {
    function CVoid(scope, node) {
        this.expression = template_1.CodeTemplateFactory.createForNode(scope, node.expression);
        scope.root.headerFlags.js_var = true;
        scope.root.headerFlags.js_var_to_undefined = true;
    }
    CVoid = __decorate([
        template_1.CodeTemplate("js_var_to_undefined({expression})", ts.SyntaxKind.VoidExpression)
    ], CVoid);
    return CVoid;
}());
var CDelete = /** @class */ (function () {
    function CDelete(scope, node) {
        this.topExpressionOfStatement = node.parent.kind == ts.SyntaxKind.ExpressionStatement;
        this.dict = (ts.isPropertyAccessExpression(node.expression) || ts.isElementAccessExpression(node.expression))
            && template_1.CodeTemplateFactory.createForNode(scope, node.expression.expression);
        if (ts.isElementAccessExpression(node.expression))
            this.argExpression = ts.isNumericLiteral(node.expression.argumentExpression)
                ? '"' + node.expression.argumentExpression.text + '"'
                : template_1.CodeTemplateFactory.createForNode(scope, node.expression.argumentExpression);
        else if (ts.isPropertyAccessExpression(node.expression))
            this.argExpression = new literals_1.CString(scope, node.expression.name.text);
        this.tempVarName = scope.root.symbolsHelper.addTemp(node, "tmp_dict_pos");
        scope.variables.push(new variable_1.CVariable(scope, this.tempVarName, types_1.NumberVarType));
        scope.root.headerFlags.bool = true;
        scope.root.headerFlags.array_remove = true;
    }
    CDelete = __decorate([
        template_1.CodeTemplate("\n{#statements}\n    {tempVarName} = dict_find_pos({dict}->index->data, {dict}->index->size, {argExpression});\n    if ({tempVarName} >= 0)\n    {\n        ARRAY_REMOVE({dict}->index, {tempVarName}, 1);\n        ARRAY_REMOVE({dict}->values, {tempVarName}, 1);\n    }\n{/statements}\n{#if !topExpressionOfStatement}\n    TRUE\n{/if}", ts.SyntaxKind.DeleteExpression)
    ], CDelete);
    return CDelete;
}());
//# sourceMappingURL=expressions.js.map