"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var template_1 = require("../template");
var variable_1 = require("./variable");
var standard_1 = require("../standard");
var typeguards_1 = require("../typeguards");
var CFunctionPrototype = /** @class */ (function () {
    function CFunctionPrototype(scope, node) {
        this.parameters = [];
        var funcType = scope.root.typeHelper.getCType(node);
        this.returnType = scope.root.typeHelper.getTypeString(funcType.returnType);
        this.name = node.name.getText();
        this.parameters = node.parameters.map(function (p, i) { return new variable_1.CVariable(scope, p.name.getText(), funcType.parameterTypes[i], { removeStorageSpecifier: true }); });
        if (funcType.instanceType)
            this.parameters.unshift(new variable_1.CVariable(scope, "this", funcType.instanceType, { removeStorageSpecifier: true }));
        for (var _i = 0, _a = funcType.closureParams; _i < _a.length; _i++) {
            var p = _a[_i];
            this.parameters.push(new variable_1.CVariable(scope, p.node.text, p.node, { removeStorageSpecifier: true }));
        }
    }
    CFunctionPrototype = __decorate([
        template_1.CodeTemplate("{returnType} {name}({parameters {, }=> {this}});")
    ], CFunctionPrototype);
    return CFunctionPrototype;
}());
exports.CFunctionPrototype = CFunctionPrototype;
var CFunction = /** @class */ (function () {
    function CFunction(root, node) {
        var _this = this;
        this.root = root;
        this.func = this;
        this.parameters = [];
        this.variables = [];
        this.statements = [];
        this.parent = root;
        this.name = node.name && node.name.text;
        if (!this.name) {
            var funcExprName = "func";
            if (typeguards_1.isEqualsExpression(node.parent) && node.parent.right == node && ts.isIdentifier(node.parent.left))
                funcExprName = node.parent.left.text + "_func";
            if (ts.isVariableDeclaration(node.parent) && node.parent.initializer == node && ts.isIdentifier(node.parent.name))
                funcExprName = node.parent.name.text + "_func";
            this.name = root.symbolsHelper.addTemp(node, funcExprName);
        }
        var funcType = root.typeHelper.getCType(node);
        this.funcDecl = new variable_1.CVariable(this, this.name, funcType.returnType, { removeStorageSpecifier: true, arraysToPointers: true });
        this.parameters = node.parameters.map(function (p, i) {
            return new variable_1.CVariable(_this, p.name.text, funcType.parameterTypes[i], { removeStorageSpecifier: true });
        });
        if (funcType.instanceType)
            this.parameters.unshift(new variable_1.CVariable(this, "this", funcType.instanceType, { removeStorageSpecifier: true }));
        for (var _i = 0, _a = funcType.closureParams; _i < _a.length; _i++) {
            var p = _a[_i];
            var type = root.typeHelper.getCType(p.node);
            this.parameters.push(new variable_1.CVariable(this, p.node.text, type, { removeStorageSpecifier: true }));
        }
        this.variables = [];
        this.gcVarNames = root.memoryManager.getGCVariablesForScope(node);
        var _loop_1 = function (gcVarName) {
            if (root.variables.filter(function (v) { return v.name == gcVarName; }).length)
                return "continue";
            var gcType = gcVarName.indexOf("arrays") == -1 ? "ARRAY(void *)" : "ARRAY(ARRAY(void *))";
            root.variables.push(new variable_1.CVariable(root, gcVarName, gcType));
        };
        for (var _b = 0, _c = this.gcVarNames; _b < _c.length; _b++) {
            var gcVarName = _c[_b];
            _loop_1(gcVarName);
        }
        node.body.statements.forEach(function (s) { return _this.statements.push(template_1.CodeTemplateFactory.createForNode(_this, s)); });
        if (node.body.statements[node.body.statements.length - 1].kind != ts.SyntaxKind.ReturnStatement) {
            this.destructors = new variable_1.CVariableDestructors(this, node);
        }
        var nodesInFunction = template_1.getAllNodesUnder(node);
        var declaredFunctionNames = root.functions.concat(root.functionPrototypes).map(function (f) { return f.name; });
        nodesInFunction.filter(function (n) { return ts.isCallExpression(n) && !standard_1.StandardCallHelper.isStandardCall(root.typeHelper, n); })
            .forEach(function (c) {
            if (ts.isIdentifier(c.expression) && declaredFunctionNames.indexOf(c.expression.text) === -1) {
                var decl = root.typeHelper.getDeclaration(c.expression);
                if (decl && decl !== node && ts.isFunctionDeclaration(decl)) {
                    root.functionPrototypes.push(new CFunctionPrototype(root, decl));
                    declaredFunctionNames.push(decl.name.text);
                }
            }
        });
    }
    CFunction = __decorate([
        template_1.CodeTemplate("\n{funcDecl}({parameters {, }=> {this}})\n{\n    {variables  {    }=> {this};\n}\n    {gcVarNames {    }=> ARRAY_CREATE({this}, 2, 0);\n}\n\n    {statements {    }=> {this}}\n\n    {destructors}\n}")
    ], CFunction);
    return CFunction;
}());
exports.CFunction = CFunction;
var CFunctionDeclaration = /** @class */ (function () {
    function CFunctionDeclaration(scope, node) {
        scope.root.functions.push(new CFunction(scope.root, node));
    }
    CFunctionDeclaration = __decorate([
        template_1.CodeTemplate("", ts.SyntaxKind.FunctionDeclaration)
    ], CFunctionDeclaration);
    return CFunctionDeclaration;
}());
exports.CFunctionDeclaration = CFunctionDeclaration;
var CFunctionExpression = /** @class */ (function () {
    function CFunctionExpression(scope, node) {
        var dynamicFunction = new CFunction(scope.root, node);
        scope.root.functions.push(dynamicFunction);
        this.name = dynamicFunction.name;
    }
    CFunctionExpression = __decorate([
        template_1.CodeTemplate("{name}", ts.SyntaxKind.FunctionExpression)
    ], CFunctionExpression);
    return CFunctionExpression;
}());
exports.CFunctionExpression = CFunctionExpression;
//# sourceMappingURL=function.js.map